#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

from abc import ABC, abstractmethod
from dataclasses import dataclass

from airbyte_cdk.sources.declarative.types import Record


@dataclass
class FileUploader(ABC):
    """
    Base class for file uploader
    """

    @abstractmethod
    def upload(self, record: Record) -> None:
        """
        Uploads the file to the specified location
        """
        ...



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Base class for source test suites."""

from dataclasses import asdict
from typing import TYPE_CHECKING

import pytest

from airbyte_cdk.models import (
    AirbyteMessage,
    AirbyteStream,
    ConfiguredAirbyteCatalog,
    ConfiguredAirbyteStream,
    DestinationSyncMode,
    SyncMode,
    Type,
)
from airbyte_cdk.test.models import (
    ConnectorTestScenario,
)
from airbyte_cdk.test.standard_tests._job_runner import run_test_job
from airbyte_cdk.test.standard_tests.connector_base import (
    ConnectorTestSuiteBase,
)

if TYPE_CHECKING:
    from airbyte_cdk.test import entrypoint_wrapper


class SourceTestSuiteBase(ConnectorTestSuiteBase):
    """Base class for source test suites.

    This class provides a base set of functionality for testing source connectors, and it
    inherits all generic connector tests from the `ConnectorTestSuiteBase` class.
    """

    def test_check(
        self,
        scenario: ConnectorTestScenario,
    ) -> None:
        """Run standard `check` tests on the connector.

        Assert that the connector returns a single CONNECTION_STATUS message.
        This test is designed to validate the connector's ability to establish a connection
        and return its status with the expected message type.
        """
        result: entrypoint_wrapper.EntrypointOutput = run_test_job(
            self.create_connector(scenario),
            "check",
            test_scenario=scenario,
            connector_root=self.get_connector_root_dir(),
        )
        num_status_messages = len(result.connection_status_messages)
        assert num_status_messages == 1, (
            f"Expected exactly one CONNECTION_STATUS message. Got {num_status_messages}: \n"
            + "\n".join([str(m) for m in result.get_message_iterator()])
        )

    def test_discover(
        self,
        scenario: ConnectorTestScenario,
    ) -> None:
        """Standard test for `discover`."""
        if scenario.expected_outcome.expect_exception():
            # If the scenario expects an exception, we can't ensure it specifically would fail
            # in discover, because some discover implementations do not need to make a connection.
            # We skip this test in that case.
            pytest.skip("Skipping discover test for scenario that expects an exception.")
            return

        run_test_job(
            self.create_connector(scenario),
            "discover",
            connector_root=self.get_connector_root_dir(),
            test_scenario=scenario,
        )

    def test_spec(self) -> None:
        """Standard test for `spec`.

        This test does not require a `scenario` input, since `spec`
        does not require any inputs.

        We assume `spec` should always succeed and it should always generate
        a valid `SPEC` message.

        Note: the parsing of messages by type also implicitly validates that
        the generated `SPEC` message is valid JSON.
        """
        result = run_test_job(
            verb="spec",
            test_scenario=None,
            connector=self.create_connector(scenario=None),
            connector_root=self.get_connector_root_dir(),
        )
        # If an error occurs, it will be raised above.

        assert len(result.spec_messages) == 1, (
            "Expected exactly 1 spec message but got {len(result.spec_messages)}",
            result.errors,
        )

    def test_basic_read(
        self,
        scenario: ConnectorTestScenario,
    ) -> None:
        """Run standard `read` test on the connector.

        This test is designed to validate the connector's ability to read data
        from the source and return records. It first runs a `discover` job to
        obtain the catalog of streams, and then it runs a `read` job to fetch
        records from those streams.
        """
        discover_result = run_test_job(
            self.create_connector(scenario),
            "discover",
            connector_root=self.get_connector_root_dir(),
            test_scenario=scenario.without_expected_outcome(),
        )
        if scenario.expected_outcome.expect_exception() and discover_result.errors:
            # Failed as expected; we're done.
            return
        streams = discover_result.catalog.catalog.streams  # type: ignore [reportOptionalMemberAccess, union-attr]

        if scenario.empty_streams:
            # Filter out streams marked as empty in the scenario.
            empty_stream_names = [stream.name for stream in scenario.empty_streams]
            streams = [s for s in streams if s.name not in empty_stream_names]

        configured_catalog = ConfiguredAirbyteCatalog(
            streams=[
                ConfiguredAirbyteStream(
                    stream=stream,
                    sync_mode=SyncMode.full_refresh,
                    destination_sync_mode=DestinationSyncMode.append_dedup,
                )
                for stream in streams
            ]
        )
        result = run_test_job(
            self.create_connector(scenario),
            "read",
            test_scenario=scenario,
            connector_root=self.get_connector_root_dir(),
            catalog=configured_catalog,
        )

        if scenario.expected_outcome.expect_success() and not result.records:
            raise AssertionError("Expected records but got none.")

    def test_fail_read_with_bad_catalog(
        self,
        scenario: ConnectorTestScenario,
    ) -> None:
        """Standard test for `read` when passed a bad catalog file."""
        invalid_configured_catalog = ConfiguredAirbyteCatalog(
            streams=[
                # Create ConfiguredAirbyteStream which is deliberately invalid
                # with regard to the Airbyte Protocol.
                # This should cause the connector to fail.
                ConfiguredAirbyteStream(
                    stream=AirbyteStream(
                        name="__AIRBYTE__stream_that_does_not_exist",
                        json_schema={
                            "type": "object",
                            "properties": {"f1": {"type": "string"}},
                        },
                        supported_sync_modes=[SyncMode.full_refresh],
                    ),
                    sync_mode="INVALID",  # type: ignore [reportArgumentType]
                    destination_sync_mode="INVALID",  # type: ignore [reportArgumentType]
                ),
            ],
        )
        result: entrypoint_wrapper.EntrypointOutput = run_test_job(
            self.create_connector(scenario),
            "read",
            connector_root=self.get_connector_root_dir(),
            test_scenario=scenario.with_expecting_failure(),  # Expect failure due to bad catalog
            catalog=asdict(invalid_configured_catalog),
        )
        assert result.errors, "Expected errors but got none."
        assert result.trace_messages, "Expected trace messages but got none."



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

from pathlib import Path

from .file_writer import FileWriter


class LocalFileSystemFileWriter(FileWriter):
    def write(self, file_path: Path, content: bytes) -> int:
        """
        Writes the file to the specified location
        """
        with open(str(file_path), "wb") as f:
            f.write(content)

        return file_path.stat().st_size



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#


from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from typing import Any, Dict

ManifestType = Dict[str, Any]


TYPE_TAG = "type"

NON_MIGRATABLE_TYPES = [
    # more info here: https://github.com/airbytehq/airbyte-internal-issues/issues/12423
    "DynamicDeclarativeStream",
]


@dataclass
class MigrationTrace:
    """
    This class represents a migration that has been applied to the manifest.
    It contains information about the migration, including the version it was applied from,
    the version it was applied to, and the time it was applied.
    """

    from_version: str
    to_version: str
    migration: str
    migrated_at: str

    def as_dict(self) -> Dict[str, Any]:
        return asdict(self)


class ManifestMigration(ABC):
    """
    Base class for manifest migrations.
    This class provides a framework for migrating manifest components.
    It defines the structure for migration classes, including methods for checking if a migration is needed,
    performing the migration, and validating the migration.
    """

    def __init__(self) -> None:
        self.is_migrated: bool = False

    @abstractmethod
    def should_migrate(self, manifest: ManifestType) -> bool:
        """
        Check if the manifest should be migrated.

        :param manifest: The manifest to potentially migrate

        :return: true if the manifest is of the expected format and should be migrated. False otherwise.
        """

    @abstractmethod
    def migrate(self, manifest: ManifestType) -> None:
        """
        Migrate the manifest. Assumes should_migrate(manifest) returned True.

        :param manifest: The manifest to migrate
        """

    @abstractmethod
    def validate(self, manifest: ManifestType) -> bool:
        """
        Validate the manifest to ensure the migration was successfully applied.

        :param manifest: The manifest to validate
        """

    def _is_component(self, obj: Dict[str, Any]) -> bool:
        """
        Check if the object is a component.

        :param obj: The object to check
        :return: True if the object is a component, False otherwise
        """
        return TYPE_TAG in obj.keys()

    def _is_migratable_type(self, obj: Dict[str, Any]) -> bool:
        """
        Check if the object is a migratable component,
        based on the Type of the component and the migration version.

        :param obj: The object to check
        :return: True if the object is a migratable component, False otherwise
        """
        return obj[TYPE_TAG] not in NON_MIGRATABLE_TYPES

    def _process_manifest(self, obj: Any) -> None:
        """
        Recursively processes a manifest object, migrating components that match the migration criteria.

        This method traverses the entire manifest structure (dictionaries and lists) and applies
        migrations to components that:
        1. Have a type tag
        2. Are not in the list of non-migratable types
        3. Meet the conditions defined in the should_migrate method

        Parameters:
            obj (Any): The object to process, which can be a dictionary, list, or any other type.
                       Dictionary objects are checked for component type tags and potentially migrated.
                       List objects have each of their items processed recursively.
                       Other types are ignored.

        Returns:
            None, since we process the manifest in place.
        """
        if isinstance(obj, dict):
            # Check if the object is a component
            if self._is_component(obj):
                # Check if the object is allowed to be migrated
                if not self._is_migratable_type(obj):
                    return

                # Check if the object should be migrated
                if self.should_migrate(obj):
                    # Perform the migration, if needed
                    self.migrate(obj)
                    # validate the migration
                    self.is_migrated = self.validate(obj)

            # Process all values in the dictionary
            for value in list(obj.values()):
                self._process_manifest(value)

        elif isinstance(obj, list):
            # Process all items in the list
            for item in obj:
                self._process_manifest(item)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import time
from collections.abc import Mapping as ABCMapping
from typing import Any, Mapping, Optional

from airbyte_cdk.models import (
    AirbyteLogMessage,
    AirbyteMessage,
    AirbyteRecordMessage,
    AirbyteRecordMessageFileReference,
    AirbyteTraceMessage,
)
from airbyte_cdk.models import Type as MessageType
from airbyte_cdk.sources.streams.core import StreamData
from airbyte_cdk.sources.utils.transform import TransformConfig, TypeTransformer


def stream_data_to_airbyte_message(
    stream_name: str,
    data_or_message: StreamData,
    transformer: TypeTransformer = TypeTransformer(TransformConfig.NoTransform),
    schema: Optional[Mapping[str, Any]] = None,
    file_reference: Optional[AirbyteRecordMessageFileReference] = None,
) -> AirbyteMessage:
    if schema is None:
        schema = {}

    match data_or_message:
        case ABCMapping():
            data = dict(data_or_message)
            now_millis = time.time_ns() // 1_000_000
            # Transform object fields according to config. Most likely you will
            # need it to normalize values against json schema. By default no action
            # taken unless configured. See
            # docs/connector-development/cdk-python/schemas.md for details.
            transformer.transform(data, schema)
            message = AirbyteRecordMessage(
                stream=stream_name,
                data=data,
                emitted_at=now_millis,
                file_reference=file_reference,
            )
            return AirbyteMessage(type=MessageType.RECORD, record=message)
        case AirbyteTraceMessage():
            return AirbyteMessage(type=MessageType.TRACE, trace=data_or_message)
        case AirbyteLogMessage():
            return AirbyteMessage(type=MessageType.LOG, log=data_or_message)
        case _:
            raise ValueError(
                f"Unexpected type for data_or_message: {type(data_or_message)}: {data_or_message}"
            )



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

"""
The AirbyteEntrypoint is important because it is a service layer that orchestrate how we execute commands from the
[common interface](https://docs.airbyte.com/understanding-airbyte/airbyte-protocol#common-interface) through the source Python
implementation. There is some logic about which message we send to the platform and when which is relevant for integration testing. Other
than that, there are integrations point that are annoying to integrate with using Python code:
* Sources communicate with the platform using stdout. The implication is that the source could just print every message instead of
    returning things to source.<method> or to using the message repository. WARNING: As part of integration testing, we will not support
    messages that are simply printed. The reason is that capturing stdout relies on overriding sys.stdout (see
    https://docs.python.org/3/library/contextlib.html#contextlib.redirect_stdout) which clashes with how pytest captures logs and brings
    considerations for multithreaded applications. If code you work with uses `print` statements, please migrate to
    source.message_repository to emit those messages
* The entrypoint interface relies on file being written on the file system
"""

import json
import logging
import re
import tempfile
import traceback
from collections import deque
from collections.abc import Generator, Mapping
from dataclasses import dataclass
from io import StringIO
from pathlib import Path
from typing import Any, List, Literal, Optional, Union, final, overload

import orjson
from pydantic import ValidationError as V2ValidationError
from serpyco_rs import SchemaValidationError

from airbyte_cdk.entrypoint import AirbyteEntrypoint
from airbyte_cdk.exception_handler import assemble_uncaught_exception
from airbyte_cdk.logger import AirbyteLogFormatter
from airbyte_cdk.models import (
    AirbyteLogMessage,
    AirbyteMessage,
    AirbyteMessageSerializer,
    AirbyteStateMessage,
    AirbyteStateMessageSerializer,
    AirbyteStreamState,
    AirbyteStreamStatus,
    ConfiguredAirbyteCatalog,
    ConfiguredAirbyteCatalogSerializer,
    Level,
    TraceType,
    Type,
)
from airbyte_cdk.sources import Source
from airbyte_cdk.test.models.scenario import ExpectedOutcome


class AirbyteEntrypointException(Exception):
    """Exception raised for errors in the AirbyteEntrypoint execution.

    Used to provide details of an Airbyte connector execution failure in the output
    captured in an `EntrypointOutput` object. Use `EntrypointOutput.as_exception()` to
    convert it to an exception.

    Example Usage:
        output = EntrypointOutput(...)
        if output.errors:
            raise output.as_exception()
    """

    message: str = ""

    def __post_init__(self) -> None:
        super().__init__(self.message)


class EntrypointOutput:
    """A class to encapsulate the output of an Airbyte connector's execution.

    This class can be initialized with a list of messages or a file containing messages.
    It provides methods to access different types of messages produced during the execution
    of an Airbyte connector, including both successful messages and error messages.

    When working with records and state messages, it provides both a list and an iterator
    implementation. Lists are easier to work with, but generators are better suited to handle
    large volumes of messages without overflowing the available memory.
    """

    def __init__(
        self,
        messages: list[str] | None = None,
        uncaught_exception: Optional[BaseException] = None,
        *,
        command: list[str] | None = None,
        message_file: Path | None = None,
    ) -> None:
        if messages is None and message_file is None:
            raise ValueError("Either messages or message_file must be provided")
        if messages is not None and message_file is not None:
            raise ValueError("Only one of messages or message_file can be provided")

        self._command = command
        self._messages: list[AirbyteMessage] | None = None
        self._message_file: Path | None = message_file
        if messages:
            try:
                self._messages = [self._parse_message(message) for message in messages]
            except V2ValidationError as exception:
                raise ValueError("All messages are expected to be AirbyteMessage") from exception

        if uncaught_exception:
            if self._messages is None:
                self._messages = []

            self._messages.append(
                assemble_uncaught_exception(
                    type(uncaught_exception), uncaught_exception
                ).as_airbyte_message()
            )

    @staticmethod
    def _parse_message(message: str) -> AirbyteMessage:
        try:
            return AirbyteMessageSerializer.load(orjson.loads(message))
        except (orjson.JSONDecodeError, SchemaValidationError):
            # The platform assumes that logs that are not of AirbyteMessage format are log messages
            return AirbyteMessage(
                type=Type.LOG, log=AirbyteLogMessage(level=Level.INFO, message=message)
            )

    @property
    def records_and_state_messages(
        self,
    ) -> list[AirbyteMessage]:
        return self.get_message_by_types(
            message_types=[Type.RECORD, Type.STATE],
            safe_iterator=False,
        )

    def records_and_state_messages_iterator(
        self,
    ) -> Generator[AirbyteMessage, None, None]:
        """Returns a generator that yields record and state messages one by one.

        Use this instead of `records_and_state_messages` when the volume of messages could be large
        enough to overload available memory.
        """
        return self.get_message_by_types(
            message_types=[Type.RECORD, Type.STATE],
            safe_iterator=True,
        )

    @property
    def records(self) -> List[AirbyteMessage]:
        return self.get_message_by_types([Type.RECORD])

    @property
    def records_iterator(self) -> Generator[AirbyteMessage, None, None]:
        """Returns a generator that yields record messages one by one.

        Use this instead of `records` when the volume of records could be large
        enough to overload available memory.
        """
        return self.get_message_by_types([Type.RECORD], safe_iterator=True)

    @property
    def state_messages(self) -> List[AirbyteMessage]:
        return self.get_message_by_types([Type.STATE])

    @property
    def spec_messages(self) -> List[AirbyteMessage]:
        return self.get_message_by_types([Type.SPEC])

    @property
    def connection_status_messages(self) -> List[AirbyteMessage]:
        return self.get_message_by_types([Type.CONNECTION_STATUS])

    @property
    def most_recent_state(self) -> AirbyteStreamState | None:
        state_message_iterator = self.get_message_by_types(
            [Type.STATE],
            safe_iterator=True,
        )
        # Use a deque with maxlen=1 to efficiently get the last state message
        double_ended_queue = deque(state_message_iterator, maxlen=1)
        try:
            final_state_message: AirbyteMessage = double_ended_queue.pop()
        except IndexError:
            raise ValueError(
                "Can't provide most recent state as there are no state messages."
            ) from None

        return final_state_message.state.stream  # type: ignore[union-attr] # state has `stream`

    @property
    def logs(self) -> List[AirbyteMessage]:
        return self.get_message_by_types([Type.LOG])

    @property
    def trace_messages(self) -> List[AirbyteMessage]:
        return self.get_message_by_types([Type.TRACE])

    @property
    def analytics_messages(self) -> List[AirbyteMessage]:
        return self._get_trace_message_by_trace_type(TraceType.ANALYTICS)

    @property
    def errors(self) -> List[AirbyteMessage]:
        return self._get_trace_message_by_trace_type(TraceType.ERROR)

    def get_formatted_error_message(self) -> str:
        """Returns a human-readable error message with the contents.

        If there are no errors, returns an empty string.
        """
        errors = self.errors
        if not errors:
            # If there are no errors, return an empty string.
            return ""

        result = "Failed to run airbyte command"
        result += ": " + " ".join(self._command) if self._command else "."
        result += "\n" + "\n".join(
            [str(error.trace.error).replace("\\n", "\n") for error in errors if error.trace],
        )
        return result

    def as_exception(self) -> AirbyteEntrypointException:
        """Convert the output to an exception."""
        return AirbyteEntrypointException(self.get_formatted_error_message())

    def raise_if_errors(
        self,
    ) -> None:
        """Raise an exception if there are errors in the output.

        Otherwise, do nothing.
        """
        if not self.errors:
            return None

        raise self.as_exception()

    @property
    def catalog(self) -> AirbyteMessage:
        catalog = self.get_message_by_types([Type.CATALOG])
        if len(catalog) != 1:
            raise ValueError(f"Expected exactly one catalog but got {len(catalog)}")
        return catalog[0]

    def get_stream_statuses(self, stream_name: str) -> List[AirbyteStreamStatus]:
        status_messages = map(
            lambda message: message.trace.stream_status.status,  # type: ignore
            filter(
                lambda message: message.trace.stream_status.stream_descriptor.name == stream_name,  # type: ignore # callable; trace has `stream_status`
                self._get_trace_message_by_trace_type(TraceType.STREAM_STATUS),
            ),
        )
        return list(status_messages)

    def get_message_iterator(self) -> Generator[AirbyteMessage, None, None]:
        """Creates a generator which yields messages one by one.

        This will iterate over all messages in the output file (if provided) or the messages
        provided during initialization. File results are provided first, followed by any
        messages that were passed in directly.
        """
        if self._message_file:
            try:
                with open(self._message_file, "r", encoding="utf-8") as file:
                    for line in file:
                        if not line.strip():
                            # Skip empty lines
                            continue

                        yield self._parse_message(line.strip())
            except FileNotFoundError:
                raise ValueError(f"Message file {self._message_file} not found")

        if self._messages is not None:
            yield from self._messages

    # Overloads to provide proper type hints for different usages of `get_message_by_types`.

    @overload
    def get_message_by_types(
        self,
        message_types: list[Type],
    ) -> list[AirbyteMessage]: ...

    @overload
    def get_message_by_types(
        self,
        message_types: list[Type],
        *,
        safe_iterator: Literal[False],
    ) -> list[AirbyteMessage]: ...

    @overload
    def get_message_by_types(
        self,
        message_types: list[Type],
        *,
        safe_iterator: Literal[True],
    ) -> Generator[AirbyteMessage, None, None]: ...

    def get_message_by_types(
        self,
        message_types: list[Type],
        *,
        safe_iterator: bool = False,
    ) -> list[AirbyteMessage] | Generator[AirbyteMessage, None, None]:
        """Get messages of specific types.

        If `safe_iterator` is True, returns a generator that yields messages one by one.
        If `safe_iterator` is False, returns a list of messages.

        Use `safe_iterator=True` when the volume of messages could overload the available
        memory.
        """
        message_generator = self.get_message_iterator()

        if safe_iterator:
            return (message for message in message_generator if message.type in message_types)

        return [message for message in message_generator if message.type in message_types]

    def _get_trace_message_by_trace_type(self, trace_type: TraceType) -> List[AirbyteMessage]:
        return [
            message
            for message in self.get_message_by_types(
                [Type.TRACE],
                safe_iterator=True,
            )
            if message.trace.type == trace_type  # type: ignore[union-attr] # trace has `type`
        ]

    def is_in_logs(self, pattern: str) -> bool:
        """Check if any log message case-insensitive matches the pattern."""
        return any(
            re.search(
                pattern,
                entry.log.message,  # type: ignore[union-attr] # log has `message`
                flags=re.IGNORECASE,
            )
            for entry in self.logs
        )

    def is_not_in_logs(self, pattern: str) -> bool:
        """Check if no log message matches the case-insensitive pattern."""
        return not self.is_in_logs(pattern)


def _run_command(
    source: Source,
    args: List[str],
    expecting_exception: bool | None = None,  # Deprecated, use `expected_outcome` instead.
    *,
    expected_outcome: ExpectedOutcome | None = None,
) -> EntrypointOutput:
    """Internal function to run a command with the AirbyteEntrypoint.

    Note: Even though this function is private, some connectors do call it directly.

    Note: The `expecting_exception` arg is now deprecated in favor of the tri-state
    `expected_outcome` arg. The old argument is supported (for now) for backwards compatibility.
    """
    expected_outcome = expected_outcome or ExpectedOutcome.from_expecting_exception_bool(
        expecting_exception,
    )
    log_capture_buffer = StringIO()
    stream_handler = logging.StreamHandler(log_capture_buffer)
    stream_handler.setLevel(logging.INFO)
    stream_handler.setFormatter(AirbyteLogFormatter())
    parent_logger = logging.getLogger("")
    parent_logger.addHandler(stream_handler)

    parsed_args = AirbyteEntrypoint.parse_args(args)

    source_entrypoint = AirbyteEntrypoint(source)
    messages: list[str] = []
    uncaught_exception = None
    try:
        for message in source_entrypoint.run(parsed_args):
            messages.append(message)
    except Exception as exception:
        if expected_outcome.expect_success():
            print("Printing unexpected error from entrypoint_wrapper")
            print("".join(traceback.format_exception(None, exception, exception.__traceback__)))

        uncaught_exception = exception

    captured_logs = log_capture_buffer.getvalue().split("\n")[:-1]

    parent_logger.removeHandler(stream_handler)
    return EntrypointOutput(
        messages=messages + captured_logs,
        uncaught_exception=uncaught_exception,
    )


def discover(
    source: Source,
    config: Mapping[str, Any],
    expecting_exception: bool | None = None,  # Deprecated, use `expected_outcome` instead.
    *,
    expected_outcome: ExpectedOutcome | None = None,
) -> EntrypointOutput:
    """
    config must be json serializable
    :param expected_outcome: By default if there is an uncaught exception, the exception will be printed out. If this is expected, please
        provide `expected_outcome=ExpectedOutcome.EXPECT_FAILURE` so that the test output logs are cleaner
    """

    with tempfile.TemporaryDirectory() as tmp_directory:
        tmp_directory_path = Path(tmp_directory)
        config_file = make_file(tmp_directory_path / "config.json", config)

        return _run_command(
            source,
            ["discover", "--config", config_file, "--debug"],
            expecting_exception=expecting_exception,  # Deprecated, but still supported.
            expected_outcome=expected_outcome,
        )


def read(
    source: Source,
    config: Mapping[str, Any],
    catalog: ConfiguredAirbyteCatalog,
    state: Optional[List[AirbyteStateMessage]] = None,
    expecting_exception: bool | None = None,  # Deprecated, use `expected_outcome` instead.
    *,
    expected_outcome: ExpectedOutcome | None = None,
    debug: bool = False,
) -> EntrypointOutput:
    """
    config and state must be json serializable

    :param expected_outcome: By default if there is an uncaught exception, the exception will be printed out. If this is expected, please
        provide `expected_outcome=ExpectedOutcome.EXPECT_FAILURE` so that the test output logs are cleaner.
    """
    with tempfile.TemporaryDirectory() as tmp_directory:
        tmp_directory_path = Path(tmp_directory)
        config_file = make_file(tmp_directory_path / "config.json", config)
        catalog_file = make_file(
            tmp_directory_path / "catalog.json",
            orjson.dumps(ConfiguredAirbyteCatalogSerializer.dump(catalog)).decode(),
        )
        args = [
            "read",
            "--config",
            config_file,
            "--catalog",
            catalog_file,
        ]
        if debug:
            args.append("--debug")
        if state is not None:
            args.extend(
                [
                    "--state",
                    make_file(
                        tmp_directory_path / "state.json",
                        f"[{','.join([orjson.dumps(AirbyteStateMessageSerializer.dump(stream_state)).decode() for stream_state in state])}]",
                    ),
                ]
            )

        return _run_command(
            source,
            args,
            expecting_exception=expecting_exception,  # Deprecated, but still supported.
            expected_outcome=expected_outcome,
        )


def make_file(
    path: Path, file_contents: Optional[Union[str, Mapping[str, Any], List[Mapping[str, Any]]]]
) -> str:
    if isinstance(file_contents, str):
        path.write_text(file_contents)
    else:
        path.write_text(json.dumps(file_contents))
    return str(path)



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

import logging
from dataclasses import InitVar, dataclass
from typing import Any, List, Mapping, Tuple, Union

from airbyte_cdk.sources import Source
from airbyte_cdk.sources.declarative.checks.check_stream import evaluate_availability
from airbyte_cdk.sources.declarative.checks.connection_checker import ConnectionChecker
from airbyte_cdk.sources.streams import Stream
from airbyte_cdk.sources.streams.concurrent.abstract_stream import AbstractStream


@dataclass
class CheckDynamicStream(ConnectionChecker):
    """
    Checks the connections by checking availability of one or many dynamic streams

    Attributes:
        stream_count (int): numbers of streams to check
    """

    # TODO: Add field stream_names to check_connection for static streams
    #  https://github.com/airbytehq/airbyte-python-cdk/pull/293#discussion_r1934933483

    stream_count: int
    parameters: InitVar[Mapping[str, Any]]
    use_check_availability: bool = True

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        self._parameters = parameters

    def check_connection(
        self,
        source: Source,
        logger: logging.Logger,
        config: Mapping[str, Any],
    ) -> Tuple[bool, Any]:
        streams: List[Union[Stream, AbstractStream]] = source.streams(config=config)  # type: ignore  # this is a migration step and we expect the declarative CDK to migrate off of ConnectionChecker

        if len(streams) == 0:
            return False, f"No streams to connect to from source {source}"
        if not self.use_check_availability:
            return True, None

        try:
            for stream in streams[: min(self.stream_count, len(streams))]:
                stream_is_available, reason = evaluate_availability(stream, logger)
                if not stream_is_available:
                    message = f"Stream {stream.name} is not available: {reason}"
                    logger.warning(message)
                    return False, message
        except Exception as error:
            error_message = (
                f"Encountered an error trying to connect to stream {stream.name}. Error: {error}"
            )
            logger.error(error_message, exc_info=True)
            return False, error_message

        return True, None



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from airbyte_cdk.sources.declarative.extractors.dpath_extractor import DpathExtractor
from airbyte_cdk.sources.declarative.extractors.http_selector import HttpSelector
from airbyte_cdk.sources.declarative.extractors.record_filter import RecordFilter
from airbyte_cdk.sources.declarative.extractors.record_selector import RecordSelector
from airbyte_cdk.sources.declarative.extractors.response_to_file_extractor import (
    ResponseToFileExtractor,
)
from airbyte_cdk.sources.declarative.extractors.type_transformer import TypeTransformer

__all__ = [
    "TypeTransformer",
    "HttpSelector",
    "DpathExtractor",
    "RecordFilter",
    "RecordSelector",
    "ResponseToFileExtractor",
]



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

from airbyte_cdk.sources.types import Record


@dataclass
class RecordMergeStrategy(ABC):
    """
    Describe the interface for how records that required multiple requests to get the complete set of fields
    should be merged back into a single record.
    """

    @abstractmethod
    def get_group_key(self, record: Record) -> Optional[str]:
        pass



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#
from dataclasses import dataclass
from typing import Any, List, MutableMapping

import dpath
import dpath.exceptions

from airbyte_cdk.sources.declarative.interpolation.interpolated_boolean import InterpolatedBoolean
from airbyte_cdk.sources.declarative.transformations.config_transformations.config_transformation import (
    ConfigTransformation,
)
from airbyte_cdk.sources.types import FieldPointer


@dataclass
class ConfigRemoveFields(ConfigTransformation):
    """
    A transformation which removes fields from a config. The fields removed are designated using FieldPointers.
    During transformation, if a field or any of its parents does not exist in the config, no error is thrown.

    If an input field pointer references an item in a list (e.g: ["k", 0] in the object {"k": ["a", "b", "c"]}) then
    the object at that index is set to None rather than being entirely removed from the list.

    It's possible to remove objects nested in lists e.g: removing [".", 0, "k"] from {".": [{"k": "V"}]} results in {".": [{}]}

    Usage syntax:

    ```yaml
        config_transformations:
          - type: RemoveFields
            field_pointers:
              - ["path", "to", "field1"]
              - ["path2"]
            condition: "{{ config.some_flag }}" # Optional condition
    ```

    Attributes:
        field_pointers (List[FieldPointer]): pointers to the fields that should be removed
        condition (str): Optional condition that determines if the fields should be removed
    """

    field_pointers: List[FieldPointer]
    condition: str = ""

    def __post_init__(self) -> None:
        self._filter_interpolator = InterpolatedBoolean(condition=self.condition, parameters={})

    def transform(
        self,
        config: MutableMapping[str, Any],
    ) -> None:
        """
        Transforms a config by removing fields based on the provided field pointers.

        :param config: The user-provided configuration to be transformed
        """
        if self.condition and not self._filter_interpolator.eval(config):
            return

        for pointer in self.field_pointers:
            try:
                dpath.delete(config, pointer)
            except dpath.exceptions.PathNotFound:
                pass



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


import logging
from abc import ABC, abstractmethod
from typing import Any, Generic, Iterable, List, Mapping, Optional, TypeVar

from airbyte_cdk.connector import BaseConnector, DefaultConnectorMixin, TConfig
from airbyte_cdk.models import (
    AirbyteCatalog,
    AirbyteMessage,
    AirbyteStateMessage,
    AirbyteStateMessageSerializer,
    ConfiguredAirbyteCatalog,
    ConfiguredAirbyteCatalogSerializer,
)

TState = TypeVar("TState")
TCatalog = TypeVar("TCatalog")


class ExperimentalClassWarning(DeprecationWarning):
    pass


class BaseSource(BaseConnector[TConfig], ABC, Generic[TConfig, TState, TCatalog]):
    @abstractmethod
    def read_state(self, state_path: str) -> TState: ...

    @abstractmethod
    def read_catalog(self, catalog_path: str) -> TCatalog: ...

    @abstractmethod
    def read(
        self,
        logger: logging.Logger,
        config: TConfig,
        catalog: TCatalog,
        state: Optional[TState] = None,
    ) -> Iterable[AirbyteMessage]:
        """
        Returns a generator of the AirbyteMessages generated by reading the source with the given configuration, catalog, and state.
        """

    @abstractmethod
    def discover(self, logger: logging.Logger, config: TConfig) -> AirbyteCatalog:
        """
        Returns an AirbyteCatalog representing the available streams and fields in this integration. For example, given valid credentials to a
        Postgres database, returns an Airbyte catalog where each postgres table is a stream, and each table column is a field.
        """


class Source(
    DefaultConnectorMixin,
    BaseSource[Mapping[str, Any], List[AirbyteStateMessage], ConfiguredAirbyteCatalog],
    ABC,
):
    # can be overridden to change an input state.
    @classmethod
    def read_state(cls, state_path: str) -> List[AirbyteStateMessage]:
        """
        Retrieves the input state of a sync by reading from the specified JSON file. Incoming state can be deserialized into either
        a JSON object for legacy state input or as a list of AirbyteStateMessages for the per-stream state format. Regardless of the
        incoming input type, it will always be transformed and output as a list of AirbyteStateMessage(s).
        :param state_path: The filepath to where the stream states are located
        :return: The complete stream state based on the connector's previous sync
        """
        parsed_state_messages = []
        if state_path:
            state_obj = BaseConnector._read_json_file(state_path)
            if state_obj:
                for state in state_obj:  # type: ignore  # `isinstance(state_obj, List)` ensures that this is a list
                    parsed_message = AirbyteStateMessageSerializer.load(state)
                    if (
                        not parsed_message.stream
                        and not parsed_message.data
                        and not parsed_message.global_
                    ):
                        raise ValueError(
                            "AirbyteStateMessage should contain either a stream, global, or state field"
                        )
                    parsed_state_messages.append(parsed_message)
        return parsed_state_messages

    # can be overridden to change an input catalog
    @classmethod
    def read_catalog(cls, catalog_path: str) -> ConfiguredAirbyteCatalog:
        return ConfiguredAirbyteCatalogSerializer.load(cls._read_json_file(catalog_path))

    @property
    def name(self) -> str:
        """Source name"""
        return self.__class__.__name__



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

import logging
from dataclasses import dataclass
from typing import Any, List

import dpath.util

from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.declarative.validators.validation_strategy import ValidationStrategy
from airbyte_cdk.sources.declarative.validators.validator import Validator

logger = logging.getLogger("airbyte")


@dataclass
class DpathValidator(Validator):
    """
    Validator that extracts a value at a specific path in the input data
    and applies a validation strategy to it.
    """

    field_path: List[str]
    strategy: ValidationStrategy

    def __post_init__(self) -> None:
        self._field_path = [
            InterpolatedString.create(path, parameters={}) for path in self.field_path
        ]
        for path_index in range(len(self.field_path)):
            if isinstance(self.field_path[path_index], str):
                self._field_path[path_index] = InterpolatedString.create(
                    self.field_path[path_index], parameters={}
                )

    def validate(self, input_data: dict[str, Any]) -> None:
        """
        Extracts the value at the specified path and applies the validation strategy.

        :param input_data: Dictionary containing the data to validate
        :raises ValueError: If the path doesn't exist or validation fails
        """
        path = [path.eval({}) for path in self._field_path]

        if len(path) == 0:
            raise ValueError("Field path is empty")

        if "*" in path:
            try:
                values = dpath.values(input_data, path)
                for value in values:
                    self.strategy.validate(value)
            except KeyError as e:
                logger.info(f"Validation skipped. Key not found: {e}")
                return

        else:
            try:
                value = dpath.get(input_data, path)
                self.strategy.validate(value)
            except KeyError as e:
                logger.info(f"Validation skipped. Key not found: {e}")
                return



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

import builtins
import datetime
import re
import typing
import uuid
from typing import Optional, Union
from urllib.parse import quote_plus

import isodate
import pytz
from dateutil import parser
from isodate import parse_duration

from airbyte_cdk.sources.declarative.datetime.datetime_parser import DatetimeParser

"""
This file contains macros that can be evaluated by a `JinjaInterpolation` object
"""


def now_utc() -> datetime.datetime:
    """
    Current local date and time in UTC timezone

    Usage:
    `"{{ now_utc() }}"`
    """
    return datetime.datetime.now(datetime.timezone.utc)


def today_utc() -> datetime.date:
    """
    Current date in UTC timezone

    Usage:
    `"{{ today_utc() }}"`
    """
    return datetime.datetime.now(datetime.timezone.utc).date()


def today_with_timezone(timezone: str) -> datetime.date:
    """
    Current date in custom timezone

    :param timezone: timezone expressed as IANA keys format. Example: "Pacific/Tarawa"
    :return:
    """
    return datetime.datetime.now(tz=pytz.timezone(timezone)).date()


def timestamp(dt: Union[float, str]) -> Union[int, float]:
    """
    Converts a number or a string to a timestamp

    If dt is a number, then convert to an int
    If dt is a string, then parse it using dateutil.parser

    Usage:
    `"{{ timestamp(1658505815.223235) }}"

    :param dt: datetime to convert to timestamp
    :return: unix timestamp
    """
    if isinstance(dt, (int, float)):
        return int(dt)
    else:
        return str_to_datetime(dt).astimezone(pytz.utc).timestamp()


def str_to_datetime(s: str) -> datetime.datetime:
    """
    Converts a string to a datetime object with UTC timezone

    If the input string does not contain timezone information, UTC is assumed.
    Supports both basic date strings like "2022-01-14" and datetime strings with optional timezone
    like "2022-01-01T13:45:30+00:00".

    Usage:
    `"{{ str_to_datetime('2022-01-14') }}"`

    :param s: string to parse as datetime
    :return: datetime object in UTC timezone
    """

    parsed_date = parser.isoparse(s)
    if not parsed_date.tzinfo:
        # Assume UTC if the input does not contain a timezone
        parsed_date = parsed_date.replace(tzinfo=pytz.utc)
    return parsed_date.astimezone(pytz.utc)


def max(*args: typing.Any) -> typing.Any:
    """
    Returns biggest object of an iterable, or two or more arguments.

    max(iterable, *[, default=obj, key=func]) -> value
    max(arg1, arg2, *args, *[, key=func]) -> value

    Usage:
    `"{{ max(2,3) }}"

    With a single iterable argument, return its biggest item. The
    default keyword-only argument specifies an object to return if
    the provided iterable is empty.
    With two or more arguments, return the largest argument.
    :param args: args to compare
    :return: largest argument
    """
    return builtins.max(*args)


def min(*args: typing.Any) -> typing.Any:
    """
    Returns smallest object of an iterable, or two or more arguments.

    min(iterable, *[, default=obj, key=func]) -> value
    min(arg1, arg2, *args, *[, key=func]) -> value

    Usage:
    `"{{ min(2,3) }}"

    With a single iterable argument, return its smallest item. The
    default keyword-only argument specifies an object to return if
    the provided iterable is empty.
    With two or more arguments, return the smallest argument.
    :param args: args to compare
    :return: smallest argument
    """
    return builtins.min(*args)


def day_delta(num_days: int, format: str = "%Y-%m-%dT%H:%M:%S.%f%z") -> str:
    """
    Returns datetime of now() + num_days

    Usage:
    `"{{ day_delta(25) }}"`

    :param num_days: number of days to add to current date time
    :return: datetime formatted as RFC3339
    """
    return (
        datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=num_days)
    ).strftime(format)


def duration(datestring: str) -> Union[datetime.timedelta, isodate.Duration]:
    """
    Converts ISO8601 duration to datetime.timedelta

    Usage:
    `"{{ now_utc() - duration('P1D') }}"`
    """
    return parse_duration(datestring)


def format_datetime(
    dt: Union[str, datetime.datetime, int], format: str, input_format: Optional[str] = None
) -> str:
    """
    Converts datetime to another format

    Usage:
    `"{{ format_datetime(config.start_date, '%Y-%m-%d') }}"`

    CPython Datetime package has known bug with `stfrtime` method: '%s' formatting uses locale timezone
    https://github.com/python/cpython/issues/77169
    https://github.com/python/cpython/issues/56959
    """
    if isinstance(dt, datetime.datetime):
        return dt.strftime(format)

    if isinstance(dt, int):
        dt_datetime = DatetimeParser().parse(dt, input_format if input_format else "%s")
    else:
        dt_datetime = (
            datetime.datetime.strptime(dt, input_format) if input_format else str_to_datetime(dt)
        )
    if dt_datetime.tzinfo is None:
        dt_datetime = dt_datetime.replace(tzinfo=pytz.utc)
    return DatetimeParser().format(dt=dt_datetime, format=format)


def sanitize_url(value: str) -> str:
    """
    Sanitizes a value by via urllib.parse.quote_plus

    Usage:
    `"{{ sanitize_url('https://example.com/path?query=value') }}"`
    """
    sanitization_strategy = quote_plus
    return sanitization_strategy(value)


def camel_case_to_snake_case(value: str) -> str:
    """
     Converts CamelCase strings to snake_case format

     Usage:
    `"{{ camel_case_to_snake_case('CamelCase') }}"`
     :param value: string to convert from CamelCase to snake_case
     :return: snake_case formatted string
    """
    return re.sub(r"(?<!^)(?=[A-Z])", "_", value).lower()


def generate_uuid() -> str:
    """
    Generates a UUID4

    Usage:
    `"{{ generate_uuid() }}"`
    """
    return str(uuid.uuid4())


_macros_list = [
    now_utc,
    today_utc,
    timestamp,
    max,
    min,
    day_delta,
    duration,
    format_datetime,
    today_with_timezone,
    str_to_datetime,
    sanitize_url,
    camel_case_to_snake_case,
    generate_uuid,
]
macros = {f.__name__: f for f in _macros_list}



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from enum import Enum
from typing import Any, List, Union

from airbyte_cdk.models import AirbyteMessage, FailureType
from airbyte_cdk.utils import AirbyteTracedException


class FileBasedSourceError(Enum):
    EMPTY_STREAM = "No files were identified in the stream. This may be because there are no files in the specified container, or because your glob patterns did not match any files. Please verify that your source contains files last modified after the start_date and that your glob patterns are not overly strict."
    GLOB_PARSE_ERROR = "Error parsing glob pattern. Please refer to the glob pattern rules at https://facelessuser.github.io/wcmatch/glob/#split."
    ENCODING_ERROR = "File encoding error. The configured encoding must match file encoding."
    ERROR_CASTING_VALUE = "Could not cast the value to the expected type."
    ERROR_CASTING_VALUE_UNRECOGNIZED_TYPE = "Could not cast the value to the expected type because the type is not recognized. Valid types are null, array, boolean, integer, number, object, and string."
    ERROR_DECODING_VALUE = "Expected a JSON-decodeable value but could not decode record."
    ERROR_LISTING_FILES = "Error listing files. Please check the credentials provided in the config and verify that they provide permission to list files."
    ERROR_READING_FILE = "Error opening file. Please check the credentials provided in the config and verify that they provide permission to read files."
    ERROR_PARSING_RECORD = "Error parsing record. This could be due to a mismatch between the config's file type and the actual file type, or because the file or record is not parseable."
    ERROR_PARSING_USER_PROVIDED_SCHEMA = (
        "The provided schema could not be transformed into valid JSON Schema."
    )
    ERROR_VALIDATING_RECORD = "One or more records do not pass the schema validation policy. Please modify your input schema, or select a more lenient validation policy."
    ERROR_VALIDATION_STREAM_DISCOVERY_OPTIONS = "Only one of options 'Schemaless', 'Input Schema', 'Files To Read For Schema Discover' or 'Use First Found File For Schema Discover' can be provided at the same time."
    ERROR_PARSING_RECORD_MISMATCHED_COLUMNS = "A header field has resolved to `None`. This indicates that the CSV has more rows than the number of header fields. If you input your schema or headers, please verify that the number of columns corresponds to the number of columns in your CSV's rows."
    ERROR_PARSING_RECORD_MISMATCHED_ROWS = "A row's value has resolved to `None`. This indicates that the CSV has more columns in the header field than the number of columns in the row(s). If you input your schema or headers, please verify that the number of columns corresponds to the number of columns in your CSV's rows."
    STOP_SYNC_PER_SCHEMA_VALIDATION_POLICY = "Stopping sync in accordance with the configured validation policy. Records in file did not conform to the schema."
    NULL_VALUE_IN_SCHEMA = "Error during schema inference: no type was detected for key."
    UNRECOGNIZED_TYPE = "Error during schema inference: unrecognized type."
    SCHEMA_INFERENCE_ERROR = "Error inferring schema from files. Are the files valid?"
    INVALID_SCHEMA_ERROR = "No fields were identified for this schema. This may happen if the stream is empty. Please check your configuration to verify that there are files that match the stream's glob patterns."
    CONFIG_VALIDATION_ERROR = "Error creating stream config object."
    MISSING_SCHEMA = "Expected `json_schema` in the configured catalog but it is missing."
    UNDEFINED_PARSER = "No parser is defined for this file type."
    UNDEFINED_VALIDATION_POLICY = (
        "The validation policy defined in the config does not exist for the source."
    )


class FileBasedErrorsCollector:
    """
    The placeholder for all errors collected.
    """

    errors: List[AirbyteMessage] = []

    def yield_and_raise_collected(self) -> Any:
        if self.errors:
            # emit collected logged messages
            yield from self.errors
            # clean the collector
            self.errors.clear()
            # raising the single exception
            raise AirbyteTracedException(
                internal_message="Please check the logged errors for more information.",
                message="Some errors occured while reading from the source.",
                failure_type=FailureType.config_error,
            )

    def collect(self, logged_error: AirbyteMessage) -> None:
        self.errors.append(logged_error)


class BaseFileBasedSourceError(Exception):
    def __init__(self, error: Union[FileBasedSourceError, str], **kwargs):  # type: ignore # noqa
        if isinstance(error, FileBasedSourceError):
            error = FileBasedSourceError(error).value
        super().__init__(
            f"{error} Contact Support if you need assistance.\n{' '.join([f'{k}={v}' for k, v in kwargs.items()])}"
        )


class ConfigValidationError(BaseFileBasedSourceError):
    pass


class InvalidSchemaError(BaseFileBasedSourceError):
    pass


class MissingSchemaError(BaseFileBasedSourceError):
    pass


class NoFilesMatchingError(BaseFileBasedSourceError):
    pass


class RecordParseError(BaseFileBasedSourceError):
    pass


class ExcelCalamineParsingError(BaseFileBasedSourceError):
    """Raised when Calamine engine fails to parse an Excel file."""

    pass


class SchemaInferenceError(BaseFileBasedSourceError):
    pass


class CheckAvailabilityError(BaseFileBasedSourceError):
    pass


class UndefinedParserError(BaseFileBasedSourceError):
    pass


class StopSyncPerValidationPolicy(BaseFileBasedSourceError):
    pass


class ErrorListingFiles(BaseFileBasedSourceError):
    pass


class DuplicatedFilesError(BaseFileBasedSourceError):
    def __init__(self, duplicated_files_names: List[dict[str, List[str]]], **kwargs: Any):
        self._duplicated_files_names = duplicated_files_names
        self._stream_name: str = kwargs["stream"]
        super().__init__(self._format_duplicate_files_error_message(), **kwargs)

    def _format_duplicate_files_error_message(self) -> str:
        duplicated_files_messages = []
        for duplicated_file in self._duplicated_files_names:
            for duplicated_file_name, file_paths in duplicated_file.items():
                file_duplicated_message = (
                    f"{len(file_paths)} duplicates found for file name {duplicated_file_name}:\n\n"
                    + "".join(f"\n - {file_paths}")
                )
                duplicated_files_messages.append(file_duplicated_message)

        error_message = (
            f"ERROR: Duplicate filenames found for stream {self._stream_name}. "
            "Duplicate file names are not allowed if the Preserve Sub-Directories in File Paths option is disabled. "
            "Please remove or rename the duplicate files before attempting to re-run the sync.\n\n"
            + "\n".join(duplicated_files_messages)
        )

        return error_message

    def __repr__(self) -> str:
        """Return a string representation of the exception."""
        class_name = self.__class__.__name__
        properties_str = ", ".join(
            f"{k}={v!r}" for k, v in self.__dict__.items() if not k.startswith("_")
        )
        return f"{class_name}({properties_str})"


class CustomFileBasedException(AirbyteTracedException):
    """
    A specialized exception for file-based connectors.

    This exception is designed to bypass the default error handling in the file-based CDK, allowing the use of custom error messages.
    """

    pass


class FileSizeLimitError(CustomFileBasedException):
    pass


class EmptyFileSchemaInferenceError(AirbyteTracedException):
    pass



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.

from dataclasses import InitVar, dataclass
from typing import Any, Iterable, List, Mapping, Optional, Set, Union

from airbyte_cdk.models import ConfiguredAirbyteStream
from airbyte_cdk.sources.declarative.requesters.query_properties import (
    PropertiesFromEndpoint,
    PropertyChunking,
)
from airbyte_cdk.sources.declarative.requesters.query_properties.property_selector import (
    PropertySelector,
)
from airbyte_cdk.sources.types import Config, StreamSlice


@dataclass
class QueryProperties:
    """
    Low-code component that encompasses the behavior to inject additional property values into the outbound API
    requests. Property values can be defined statically within the manifest or dynamically by making requests
    to a partner API to retrieve the properties. Query properties also allow for splitting of the total set of
    properties into smaller chunks to satisfy API restrictions around the total amount of data retrieved
    """

    property_list: Optional[Union[List[str], PropertiesFromEndpoint]]
    always_include_properties: Optional[List[str]]
    property_chunking: Optional[PropertyChunking]
    property_selector: Optional[PropertySelector]
    config: Config
    parameters: InitVar[Mapping[str, Any]]

    def get_request_property_chunks(self) -> Iterable[List[str]]:
        """
        Uses the defined property_list to fetch the total set of properties dynamically or from a static list
        and based on the resulting properties, performs property chunking if applicable.
        """
        fields: List[str]
        configured_properties = self.property_selector.select() if self.property_selector else None

        if isinstance(self.property_list, PropertiesFromEndpoint):
            fields = self.property_list.get_properties_from_endpoint()
        else:
            fields = self.property_list if self.property_list else []

        if self.property_chunking:
            yield from self.property_chunking.get_request_property_chunks(
                property_fields=fields,
                always_include_properties=self.always_include_properties,
                configured_properties=configured_properties,
            )
        else:
            if configured_properties is not None:
                all_fields = (
                    [field for field in fields if field in configured_properties]
                    if configured_properties is not None
                    else list(fields)
                )
            else:
                all_fields = list(fields)

            if self.always_include_properties:
                all_fields = list(self.always_include_properties) + all_fields

            yield all_fields



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Manifest related commands.

Coming soon.

This module is planned to provide a command line interface (CLI) for validating
Airbyte CDK manifests.
"""

import rich_click as click


@click.group(
    name="manifest",
    help=__doc__.replace("\n", "\n\n"),  # Render docstring as help text (markdown)
)
def manifest_cli_group() -> None:
    """Manifest related commands."""
    pass


__all__ = [
    "manifest_cli_group",
]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass
from enum import Enum
from typing import Any, List, Literal, Mapping, MutableMapping, Optional, Union

from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.types import Config


class RequestOptionType(Enum):
    """
    Describes where to set a value on a request
    """

    request_parameter = "request_parameter"
    header = "header"
    body_data = "body_data"
    body_json = "body_json"


@dataclass
class RequestOption:
    """
    Describes an option to set on a request

    Attributes:
        field_name (str): Describes the name of the parameter to inject. Mutually exclusive with field_path.
        field_path (list(str)): Describes the path to a nested field as a list of field names.
          Only valid for body_json injection type, and mutually exclusive with field_name.
        inject_into (RequestOptionType): Describes where in the HTTP request to inject the parameter
    """

    inject_into: RequestOptionType
    parameters: InitVar[Mapping[str, Any]]
    field_name: Optional[Union[InterpolatedString, str]] = None
    field_path: Optional[List[Union[InterpolatedString, str]]] = None

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        # Validate inputs. We should expect either field_name or field_path, but not both
        if self.field_name is None and self.field_path is None:
            raise ValueError("RequestOption requires either a field_name or field_path")

        if self.field_name is not None and self.field_path is not None:
            raise ValueError(
                "Only one of field_name or field_path can be provided to RequestOption"
            )

        # Nested field injection is only supported for body JSON injection
        if self.field_path is not None and self.inject_into != RequestOptionType.body_json:
            raise ValueError(
                "Nested field injection is only supported for body JSON injection. Please use a top-level field_name for other injection types."
            )

        # Convert field_name and field_path into InterpolatedString objects if they are strings
        if self.field_name is not None:
            self.field_name = InterpolatedString.create(self.field_name, parameters=parameters)
        elif self.field_path is not None:
            self.field_path = [
                InterpolatedString.create(segment, parameters=parameters)
                for segment in self.field_path
            ]

    @property
    def _is_field_path(self) -> bool:
        """Returns whether this option is a field path (ie, a nested field)"""
        return self.field_path is not None

    def inject_into_request(
        self,
        target: MutableMapping[str, Any],
        value: Any,
        config: Config,
    ) -> None:
        """
        Inject a request option value into a target request structure using either field_name or field_path.
        For non-body-json injection, only top-level field names are supported.
        For body-json injection, both field names and nested field paths are supported.

        Args:
            target: The request structure to inject the value into
            value: The value to inject
            config: The config object to use for interpolation
        """
        if self._is_field_path:
            if self.inject_into != RequestOptionType.body_json:
                raise ValueError(
                    "Nested field injection is only supported for body JSON injection. Please use a top-level field_name for other injection types."
                )

            assert self.field_path is not None  # for type checker
            current = target
            # Convert path segments into strings, evaluating any interpolated segments
            # Example: ["data", "{{ config[user_type] }}", "id"] -> ["data", "admin", "id"]
            *path_parts, final_key = [
                str(
                    segment.eval(config=config)
                    if isinstance(segment, InterpolatedString)
                    else segment
                )
                for segment in self.field_path
            ]

            # Build a nested dictionary structure and set the final value at the deepest level
            for part in path_parts:
                current = current.setdefault(part, {})
            current[final_key] = value
        else:
            # For non-nested fields, evaluate the field name if it's an interpolated string
            key = (
                self.field_name.eval(config=config)
                if isinstance(self.field_name, InterpolatedString)
                else self.field_name
            )
            target[str(key)] = value



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from abc import ABC, abstractmethod
from typing import Any, Optional

import requests

from airbyte_cdk.sources.declarative.requesters.paginators.strategies.pagination_strategy import (
    PaginationStrategy,
)
from airbyte_cdk.sources.streams.concurrent.cursor import Cursor
from airbyte_cdk.sources.types import Record


class PaginationStopCondition(ABC):
    @abstractmethod
    def is_met(self, record: Record) -> bool:
        """
        Given a condition is met, the pagination will stop

        :param record: a record used to evaluate the condition
        """
        raise NotImplementedError()


class CursorStopCondition(PaginationStopCondition):
    def __init__(
        self,
        cursor: Cursor,
    ):
        self._cursor = cursor

    def is_met(self, record: Record) -> bool:
        return not self._cursor.should_be_synced(record)


class StopConditionPaginationStrategyDecorator(PaginationStrategy):
    def __init__(self, _delegate: PaginationStrategy, stop_condition: PaginationStopCondition):
        self._delegate = _delegate
        self._stop_condition = stop_condition

    def next_page_token(
        self,
        response: requests.Response,
        last_page_size: int,
        last_record: Optional[Record],
        last_page_token_value: Optional[Any] = None,
    ) -> Optional[Any]:
        # We evaluate in reverse order because the assumption is that most of the APIs using data feed structure
        # will return records in descending order. In terms of performance/memory, we return the records lazily
        if last_record and self._stop_condition.is_met(last_record):
            return None
        return self._delegate.next_page_token(
            response, last_page_size, last_record, last_page_token_value
        )

    def get_page_size(self) -> Optional[int]:
        return self._delegate.get_page_size()

    @property
    def initial_token(self) -> Optional[Any]:
        return self._delegate.initial_token



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.

# THIS IS A STATIC CLASS MODEL USED TO DISPLAY DEPRECATION WARNINGS
# WHEN DEPRECATED FIELDS ARE ACCESSED

import warnings

# ignore the SyntaxWarning in the Airbyte log messages, during the string evaluation
warnings.filterwarnings("ignore", category=SyntaxWarning)

from typing import Any, List

from pydantic.v1 import BaseModel

from airbyte_cdk.connector_builder.models import LogMessage as ConnectorBuilderLogMessage

# format the warning message
warnings.formatwarning = (
    lambda message, category, *args, **kwargs: f"{category.__name__}: {message}\n"
)


FIELDS_TAG = "__fields__"
DEPRECATED = "deprecated"
DEPRECATION_MESSAGE = "deprecation_message"
DEPRECATION_LOGS_TAG = "_deprecation_logs"


class BaseModelWithDeprecations(BaseModel):
    """
    Pydantic BaseModel that warns when deprecated fields are accessed.
    The deprecation message is stored in the field's extra attributes.
    This class is used to create models that can have deprecated fields
    and show warnings when those fields are accessed or initialized.

    The `_deprecation_logs` attribute is stored in the model itself.
    The collected deprecation warnings are further propagated to the Airbyte log messages,
    during the component creation process, in `model_to_component._collect_model_deprecations()`.

    The component implementation is not responsible for handling the deprecation warnings,
    since the deprecation warnings are already handled in the model itself.
    """

    class Config:
        """
        Allow extra fields in the model. In case the model restricts extra fields.
        """

        extra = "allow"

    def __init__(self, **model_fields: Any) -> None:
        """
        Show warnings for deprecated fields during component initialization.
        """
        # call the parent constructor first to initialize Pydantic internals
        super().__init__(**model_fields)
        # set the placeholder for the default deprecation messages
        self._default_deprecation_messages: List[str] = []
        # set the placeholder for the deprecation logs
        self._deprecation_logs: List[ConnectorBuilderLogMessage] = []
        # process deprecated fields, if present
        self._process_fields(model_fields)
        # emit default deprecation messages
        self._emit_default_deprecation_messages()
        # set the deprecation logs attribute to the model
        self._set_deprecation_logs_attr_to_model()

    def _is_deprecated_field(self, field_name: str) -> bool:
        return (
            self.__fields__[field_name].field_info.extra.get(DEPRECATED, False)
            if field_name in self.__fields__.keys()
            else False
        )

    def _get_deprecation_message(self, field_name: str) -> str:
        return (
            self.__fields__[field_name].field_info.extra.get(
                DEPRECATION_MESSAGE, "<missing_deprecation_message>"
            )
            if field_name in self.__fields__.keys()
            else "<missing_deprecation_message>"
        )

    def _process_fields(self, model_fields: Any) -> None:
        """
        Processes the fields in the provided model data, checking for deprecated fields.

        For each field in the input `model_fields`, this method checks if the field exists in the model's defined fields.
        If the field is marked as deprecated (using the `DEPRECATED` flag in its metadata), it triggers a deprecation warning
        by calling the `_create_warning` method with the field name and an optional deprecation message.

        Args:
            model_fields (Any): The data containing fields to be processed.

        Returns:
            None
        """

        if hasattr(self, FIELDS_TAG):
            for field_name in model_fields.keys():
                if self._is_deprecated_field(field_name):
                    self._create_warning(
                        field_name,
                        self._get_deprecation_message(field_name),
                    )

    def _set_deprecation_logs_attr_to_model(self) -> None:
        """
        Sets the deprecation logs attribute on the model instance.

        This method attaches the current instance's deprecation logs to the model by setting
        an attribute named by `DEPRECATION_LOGS_TAG` to the value of `self._deprecation_logs`.
        This is typically used to track or log deprecated features or configurations within the model.

        Returns:
            None
        """
        setattr(self, DEPRECATION_LOGS_TAG, self._deprecation_logs)

    def _create_warning(self, field_name: str, message: str) -> None:
        """
        Show a warning message for deprecated fields (to stdout).
        Args:
            field_name (str): Name of the deprecated field.
            message (str): Warning message to be displayed.
        """

        deprecated_message = f"Component type: `{self.__class__.__name__}`. Field '{field_name}' is deprecated. {message}"

        if deprecated_message not in self._default_deprecation_messages:
            # Avoid duplicates in the default deprecation messages
            self._default_deprecation_messages.append(deprecated_message)

        # Create an Airbyte deprecation log message
        deprecation_log_message = ConnectorBuilderLogMessage(
            level="WARN", message=deprecated_message
        )
        # Add the deprecation message to the Airbyte log messages,
        # this logs are displayed in the Connector Builder.
        if deprecation_log_message not in self._deprecation_logs:
            # Avoid duplicates in the deprecation logs
            self._deprecation_logs.append(deprecation_log_message)

    def _emit_default_deprecation_messages(self) -> None:
        """
        Emit default deprecation messages for deprecated fields to STDOUT.
        """
        for message in self._default_deprecation_messages:
            warnings.warn(message, DeprecationWarning)



#
# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
#

from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import Any, Dict, Iterable, List, Mapping

import dpath
import yaml
from typing_extensions import deprecated
from yaml.parser import ParserError

from airbyte_cdk.sources.declarative.interpolation import InterpolatedString
from airbyte_cdk.sources.declarative.resolvers.components_resolver import (
    ComponentMappingDefinition,
    ComponentsResolver,
    ResolvedComponentMappingDefinition,
)
from airbyte_cdk.sources.source import ExperimentalClassWarning
from airbyte_cdk.sources.types import Config


@deprecated("This class is experimental. Use at your own risk.", category=ExperimentalClassWarning)
@dataclass
class StreamParametersDefinition:
    """
    Represents a stream parameters definition to set up dynamic streams from defined values in manifest.
    """

    list_of_parameters_for_stream: List[Dict[str, Any]]


@deprecated("This class is experimental. Use at your own risk.", category=ExperimentalClassWarning)
@dataclass
class ParametrizedComponentsResolver(ComponentsResolver):
    """
    Resolves and populates dynamic streams from defined parametrized values in manifest.
    """

    stream_parameters: StreamParametersDefinition
    config: Config
    components_mapping: List[ComponentMappingDefinition]
    parameters: InitVar[Mapping[str, Any]]
    _resolved_components: List[ResolvedComponentMappingDefinition] = field(
        init=False, repr=False, default_factory=list
    )

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        """
        Initializes and parses component mappings, converting them to resolved definitions.

        Args:
            parameters (Mapping[str, Any]): Parameters for interpolation.
        """

        for component_mapping in self.components_mapping:
            if isinstance(component_mapping.value, (str, InterpolatedString)):
                interpolated_value = (
                    InterpolatedString.create(component_mapping.value, parameters=parameters)
                    if isinstance(component_mapping.value, str)
                    else component_mapping.value
                )

                field_path = [
                    InterpolatedString.create(path, parameters=parameters)
                    for path in component_mapping.field_path
                ]

                self._resolved_components.append(
                    ResolvedComponentMappingDefinition(
                        field_path=field_path,
                        value=interpolated_value,
                        value_type=component_mapping.value_type,
                        create_or_update=component_mapping.create_or_update,
                        parameters=parameters,
                    )
                )
            else:
                raise ValueError(
                    f"Expected a string or InterpolatedString for value in mapping: {component_mapping}"
                )

    def resolve_components(
        self, stream_template_config: Dict[str, Any]
    ) -> Iterable[Dict[str, Any]]:
        kwargs = {"stream_template_config": stream_template_config}

        for components_values in self.stream_parameters.list_of_parameters_for_stream:
            updated_config = deepcopy(stream_template_config)
            kwargs["components_values"] = components_values  # type: ignore[assignment] # component_values will always be of type Mapping[str, Any]
            for resolved_component in self._resolved_components:
                valid_types = (
                    (resolved_component.value_type,) if resolved_component.value_type else None
                )
                value = resolved_component.value.eval(
                    self.config, valid_types=valid_types, **kwargs
                )
                path = [path.eval(self.config, **kwargs) for path in resolved_component.field_path]
                parsed_value = self._parse_yaml_if_possible(value)
                # https://github.com/dpath-maintainers/dpath-python/blob/master/dpath/__init__.py#L136
                # dpath.set returns the number of changed elements, 0 when no elements changed
                updated = dpath.set(updated_config, path, parsed_value)

                if parsed_value and not updated and resolved_component.create_or_update:
                    dpath.new(updated_config, path, parsed_value)

            yield updated_config

    @staticmethod
    def _parse_yaml_if_possible(value: Any) -> Any:
        """
        Try to turn value into a Python object by YAML-parsing it.

        * If value is a `str` and can be parsed by `yaml.safe_load`,
          return the parsed result.
        * If parsing fails (`yaml.parser.ParserError`) – or value is not
          a string at all – return the original value unchanged.
        """
        if isinstance(value, str):
            try:
                return yaml.safe_load(value)
            except ParserError:  # "{{ record[0] in ['cohortActiveUsers'] }}"   # not valid YAML
                return value
        return value



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from __future__ import annotations

import logging
from abc import abstractmethod
from typing import TYPE_CHECKING, Optional, Tuple

from airbyte_cdk.sources import Source
from airbyte_cdk.sources.streams.availability_strategy import AvailabilityStrategy
from airbyte_cdk.sources.streams.core import Stream

if TYPE_CHECKING:
    from airbyte_cdk.sources.file_based.stream import AbstractFileBasedStream


class AbstractFileBasedAvailabilityStrategy(AvailabilityStrategy):
    @abstractmethod
    def check_availability(  # type: ignore[override]  # Signature doesn't match base class
        self,
        stream: Stream,
        logger: logging.Logger,
        source: Optional[Source] = None,
    ) -> Tuple[bool, Optional[str]]:
        """
        Perform a connection check for the stream.

        Returns (True, None) if successful, otherwise (False, <error message>).
        """
        ...

    @abstractmethod
    def check_availability_and_parsability(
        self,
        stream: AbstractFileBasedStream,
        logger: logging.Logger,
        _: Optional[Source],
    ) -> Tuple[bool, Optional[str]]:
        """
        Performs a connection check for the stream, as well as additional checks that
        verify that the connection is working as expected.

        Returns (True, None) if successful, otherwise (False, <error message>).
        """
        ...



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Start command for the manifest server CLI."""

import rich_click as click

from ._common import check_manifest_server_dependencies


@click.command()
@click.option(
    "--host",
    default="127.0.0.1",
    help="Host to bind the server to",
    show_default=True,
)
@click.option(
    "--port",
    default=8000,
    help="Port to bind the server to",
    show_default=True,
)
@click.option(
    "--reload",
    is_flag=True,
    help="Enable auto-reload for development",
)
def start(host: str, port: int, reload: bool) -> None:
    """Start the FastAPI manifest server server."""
    check_manifest_server_dependencies()

    # Import and use the main server function
    from airbyte_cdk.manifest_server.main import run_server

    run_server(
        host=host,
        port=port,
        reload=reload,
    )



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from logging import Logger
from typing import Any, Callable, Iterable, List, Mapping, Optional, Union

from airbyte_cdk.models import AirbyteStream, SyncMode
from airbyte_cdk.sources.streams.concurrent.abstract_stream import AbstractStream
from airbyte_cdk.sources.streams.concurrent.availability_strategy import StreamAvailability
from airbyte_cdk.sources.streams.concurrent.cursor import Cursor, CursorField
from airbyte_cdk.sources.streams.concurrent.partitions.partition import Partition
from airbyte_cdk.sources.streams.concurrent.partitions.partition_generator import PartitionGenerator
from airbyte_cdk.utils.traced_exception import AirbyteTracedException


class DefaultStream(AbstractStream):
    def __init__(
        self,
        partition_generator: PartitionGenerator,
        name: str,
        json_schema: Union[Mapping[str, Any], Callable[[], Mapping[str, Any]]],
        primary_key: List[str],
        cursor_field: Optional[CursorField],
        logger: Logger,
        cursor: Cursor,
        namespace: Optional[str] = None,
        supports_file_transfer: bool = False,
    ) -> None:
        self._stream_partition_generator = partition_generator
        self._name = name
        self._json_schema = json_schema
        self._primary_key = primary_key
        self._cursor_field = cursor_field
        self._logger = logger
        self._cursor = cursor
        self._namespace = namespace
        self._supports_file_transfer = supports_file_transfer

    def generate_partitions(self) -> Iterable[Partition]:
        yield from self._stream_partition_generator.generate()

    @property
    def name(self) -> str:
        return self._name

    @property
    def namespace(self) -> Optional[str]:
        return self._namespace

    @property
    def cursor_field(self) -> Optional[str]:
        return self._cursor_field.cursor_field_key if self._cursor_field else None

    def get_json_schema(self) -> Mapping[str, Any]:
        return self._json_schema() if callable(self._json_schema) else self._json_schema

    def as_airbyte_stream(self) -> AirbyteStream:
        stream = AirbyteStream(
            name=self.name,
            json_schema=dict(self.get_json_schema()),
            supported_sync_modes=[SyncMode.full_refresh],
            is_resumable=False,
            is_file_based=self._supports_file_transfer,
        )

        if self._namespace:
            stream.namespace = self._namespace

        if self._cursor_field:
            stream.source_defined_cursor = (
                not self._cursor_field.supports_catalog_defined_cursor_field
            )
            stream.is_resumable = True
            stream.supported_sync_modes.append(SyncMode.incremental)
            stream.default_cursor_field = [self._cursor_field.cursor_field_key]

        keys = self._primary_key
        if keys and len(keys) > 0:
            stream.source_defined_primary_key = [[key] for key in keys]

        return stream

    def log_stream_sync_configuration(self) -> None:
        self._logger.debug(
            f"Syncing stream instance: {self.name}",
            extra={
                "primary_key": self._primary_key,
                "cursor_field": self.cursor_field,
            },
        )

    @property
    def cursor(self) -> Cursor:
        return self._cursor

    def check_availability(self) -> StreamAvailability:
        """
        Check stream availability by attempting to read the first record of the stream.
        """
        try:
            partition = next(iter(self.generate_partitions()))
        except StopIteration:
            # NOTE: The following comment was copied from legacy stuff and I don't know how relevant it is:
            # If stream_slices has no `next()` item (Note - this is different from stream_slices returning [None]!)
            # This can happen when a substream's `stream_slices` method does a `for record in parent_records: yield <something>`
            # without accounting for the case in which the parent stream is empty.
            return StreamAvailability.unavailable(
                f"Cannot attempt to connect to stream {self.name} - no stream slices were found"
            )
        except AirbyteTracedException as error:
            return StreamAvailability.unavailable(
                error.message or error.internal_message or "<no error message>"
            )

        try:
            next(iter(partition.read()))
            return StreamAvailability.available()
        except StopIteration:
            self._logger.info(f"Successfully connected to stream {self.name}, but got 0 records.")
            return StreamAvailability.available()
        except AirbyteTracedException as error:
            return StreamAvailability.unavailable(
                error.message or error.internal_message or "<no error message>"
            )



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
import copy
from dataclasses import InitVar, dataclass, field
from typing import Any, List, Mapping, Optional, Set

from airbyte_protocol_dataclasses.models import ConfiguredAirbyteStream

from airbyte_cdk.sources.declarative.requesters.query_properties.property_selector.property_selector import (
    PropertySelector,
)
from airbyte_cdk.sources.declarative.transformations import RecordTransformation
from airbyte_cdk.sources.types import Config


@dataclass
class JsonSchemaPropertySelector(PropertySelector):
    """
    A class that contains a list of transformations to apply to properties.
    """

    config: Config
    parameters: InitVar[Mapping[str, Any]]
    # For other non-read operations, there is no configured catalog and therefore no schema selection
    configured_stream: Optional[ConfiguredAirbyteStream] = None
    properties_transformations: List[RecordTransformation] = field(default_factory=lambda: [])

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        self._parameters = parameters

    def select(self) -> Optional[Set[str]]:
        """
        Returns the set of properties that have been selected for the configured stream. The intent being that
        we should only query for selected properties not all since disabled properties are discarded.

        When configured_stream is None, then there was no incoming catalog and all fields should be retrieved.
        This is different from the empty set where the json_schema was empty and no schema fields were selected.
        """

        # For CHECK/DISCOVER operations, there is no catalog and therefore no configured stream or selected
        # columns. In this case we return None which is interpreted by the QueryProperties component to not
        # perform any filtering of schema properties and fetch all of them
        if self.configured_stream is None:
            return None

        schema_properties = copy.deepcopy(
            self.configured_stream.stream.json_schema.get("properties", {})
        )
        if self.properties_transformations:
            for transformation in self.properties_transformations:
                transformation.transform(
                    record=schema_properties,
                    config=self.config,
                )
        return set(schema_properties.keys())



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Mapping, Optional, Union

from airbyte_cdk.sources.types import StreamSlice, StreamState


@dataclass
class RequestOptionsProvider:
    """
    Defines the request options to set on an outgoing HTTP request

    Options can be passed by
    - request parameter
    - request headers
    - body data
    - json content
    """

    @abstractmethod
    def get_request_params(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        """
        Specifies the query parameters that should be set on an outgoing HTTP request given the inputs.

        E.g: you might want to define query parameters for paging if next_page_token is not None.
        """
        pass

    @abstractmethod
    def get_request_headers(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        """Return any non-auth headers. Authentication headers will overwrite any overlapping headers returned from this method."""

    @abstractmethod
    def get_request_body_data(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Union[Mapping[str, Any], str]:
        """
        Specifies how to populate the body of the request with a non-JSON payload.

        If returns a ready text that it will be sent as is.
        If returns a dict that it will be converted to a urlencoded form.
        E.g. {"key1": "value1", "key2": "value2"} => "key1=value1&key2=value2"

        At the same time only one of the 'request_body_data' and 'request_body_json' functions can be overridden.
        """

    @abstractmethod
    def get_request_body_json(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        """
        Specifies how to populate the body of the request with a JSON payload.

        At the same time only one of the 'request_body_data' and 'request_body_json' functions can be overridden.
        """



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#


from urllib.parse import urljoin

from airbyte_cdk.manifest_migrations.manifest_migration import (
    TYPE_TAG,
    ManifestMigration,
    ManifestType,
)
from airbyte_cdk.sources.types import EmptyString


class HttpRequesterPathToUrl(ManifestMigration):
    """
    This migration is responsible for migrating the `path` key to `url` in the HttpRequester component.
    The `path` key is expected to be a relative path, and the `url` key is expected to be a full URL.
    The migration will concatenate the `url_base` and `path` to form a full URL.
    """

    component_type = "HttpRequester"
    original_key = "path"
    replacement_key = "url"

    def should_migrate(self, manifest: ManifestType) -> bool:
        return manifest[TYPE_TAG] == self.component_type and self.original_key in list(
            manifest.keys()
        )

    def migrate(self, manifest: ManifestType) -> None:
        original_key_value = manifest.get(self.original_key, EmptyString).lstrip("/")
        replacement_key_value = manifest[self.replacement_key]

        # return a full-url if provided directly from interpolation context
        if original_key_value == EmptyString or original_key_value is None:
            manifest[self.replacement_key] = replacement_key_value
            manifest.pop(self.original_key, None)
        else:
            # since we didn't provide a full-url, the url_base might not have a trailing slash
            # so we join the url_base and path correctly
            if not replacement_key_value.endswith("/"):
                replacement_key_value += "/"

            manifest[self.replacement_key] = urljoin(replacement_key_value, original_key_value)
            manifest.pop(self.original_key, None)

    def validate(self, manifest: ManifestType) -> bool:
        """
        Validate the migration by checking if the `url` key is present and the `path` key is not.
        """
        return (
            self.replacement_key in manifest
            and self.original_key not in manifest
            and manifest[self.replacement_key] is not None
        )



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import concurrent
import logging
from queue import Queue
from typing import Iterable, Iterator, List, Optional

from airbyte_cdk.models import AirbyteMessage
from airbyte_cdk.sources.concurrent_source.concurrent_read_processor import ConcurrentReadProcessor
from airbyte_cdk.sources.concurrent_source.partition_generation_completed_sentinel import (
    PartitionGenerationCompletedSentinel,
)
from airbyte_cdk.sources.concurrent_source.stream_thread_exception import StreamThreadException
from airbyte_cdk.sources.concurrent_source.thread_pool_manager import ThreadPoolManager
from airbyte_cdk.sources.message import InMemoryMessageRepository, MessageRepository
from airbyte_cdk.sources.streams.concurrent.abstract_stream import AbstractStream
from airbyte_cdk.sources.streams.concurrent.partition_enqueuer import PartitionEnqueuer
from airbyte_cdk.sources.streams.concurrent.partition_reader import PartitionLogger, PartitionReader
from airbyte_cdk.sources.streams.concurrent.partitions.partition import Partition
from airbyte_cdk.sources.streams.concurrent.partitions.types import (
    PartitionCompleteSentinel,
    QueueItem,
)
from airbyte_cdk.sources.types import Record
from airbyte_cdk.sources.utils.slice_logger import DebugSliceLogger, SliceLogger


class ConcurrentSource:
    """
    A Source that reads data from multiple AbstractStreams concurrently.
    It does so by submitting partition generation, and partition read tasks to a thread pool.
    The tasks asynchronously add their output to a shared queue.
    The read is done when all partitions for all streams w ere generated and read.
    """

    DEFAULT_TIMEOUT_SECONDS = 900

    @staticmethod
    def create(
        num_workers: int,
        initial_number_of_partitions_to_generate: int,
        logger: logging.Logger,
        slice_logger: SliceLogger,
        message_repository: MessageRepository,
        queue: Optional[Queue[QueueItem]] = None,
        timeout_seconds: int = DEFAULT_TIMEOUT_SECONDS,
    ) -> "ConcurrentSource":
        is_single_threaded = initial_number_of_partitions_to_generate == 1 and num_workers == 1
        too_many_generator = (
            not is_single_threaded and initial_number_of_partitions_to_generate >= num_workers
        )
        assert not too_many_generator, (
            "It is required to have more workers than threads generating partitions"
        )
        threadpool = ThreadPoolManager(
            concurrent.futures.ThreadPoolExecutor(
                max_workers=num_workers, thread_name_prefix="workerpool"
            ),
            logger,
        )
        return ConcurrentSource(
            threadpool=threadpool,
            logger=logger,
            slice_logger=slice_logger,
            queue=queue,
            message_repository=message_repository,
            initial_number_partitions_to_generate=initial_number_of_partitions_to_generate,
            timeout_seconds=timeout_seconds,
        )

    def __init__(
        self,
        threadpool: ThreadPoolManager,
        logger: logging.Logger,
        slice_logger: SliceLogger = DebugSliceLogger(),
        queue: Optional[Queue[QueueItem]] = None,
        message_repository: MessageRepository = InMemoryMessageRepository(),
        initial_number_partitions_to_generate: int = 1,
        timeout_seconds: int = DEFAULT_TIMEOUT_SECONDS,
    ) -> None:
        """
        :param threadpool: The threadpool to submit tasks to
        :param logger: The logger to log to
        :param slice_logger: The slice logger used to create messages on new slices
        :param message_repository: The repository to emit messages to
        :param initial_number_partitions_to_generate: The initial number of concurrent partition generation tasks. Limiting this number ensures will limit the latency of the first records emitted. While the latency is not critical, emitting the records early allows the platform and the destination to process them as early as possible.
        :param timeout_seconds: The maximum number of seconds to wait for a record to be read from the queue. If no record is read within this time, the source will stop reading and return.
        """
        self._threadpool = threadpool
        self._logger = logger
        self._slice_logger = slice_logger
        self._message_repository = message_repository
        self._initial_number_partitions_to_generate = initial_number_partitions_to_generate
        self._timeout_seconds = timeout_seconds

        # We set a maxsize to for the main thread to process record items when the queue size grows. This assumes that there are less
        # threads generating partitions that than are max number of workers. If it weren't the case, we could have threads only generating
        # partitions which would fill the queue. This number is arbitrarily set to 10_000 but will probably need to be changed given more
        # information and might even need to be configurable depending on the source
        self._queue = queue or Queue(maxsize=10_000)

    def read(
        self,
        streams: List[AbstractStream],
    ) -> Iterator[AirbyteMessage]:
        self._logger.info("Starting syncing")
        concurrent_stream_processor = ConcurrentReadProcessor(
            streams,
            PartitionEnqueuer(self._queue, self._threadpool),
            self._threadpool,
            self._logger,
            self._slice_logger,
            self._message_repository,
            PartitionReader(
                self._queue,
                PartitionLogger(self._slice_logger, self._logger, self._message_repository),
            ),
        )

        # Enqueue initial partition generation tasks
        yield from self._submit_initial_partition_generators(concurrent_stream_processor)

        # Read from the queue until all partitions were generated and read
        yield from self._consume_from_queue(
            self._queue,
            concurrent_stream_processor,
        )
        self._threadpool.check_for_errors_and_shutdown()
        self._logger.info("Finished syncing")

    def _submit_initial_partition_generators(
        self, concurrent_stream_processor: ConcurrentReadProcessor
    ) -> Iterable[AirbyteMessage]:
        for _ in range(self._initial_number_partitions_to_generate):
            status_message = concurrent_stream_processor.start_next_partition_generator()
            if status_message:
                yield status_message

    def _consume_from_queue(
        self,
        queue: Queue[QueueItem],
        concurrent_stream_processor: ConcurrentReadProcessor,
    ) -> Iterable[AirbyteMessage]:
        while airbyte_message_or_record_or_exception := queue.get():
            yield from self._handle_item(
                airbyte_message_or_record_or_exception,
                concurrent_stream_processor,
            )
            # In the event that a partition raises an exception, anything remaining in
            # the queue will be missed because is_done() can raise an exception and exit
            # out of this loop before remaining items are consumed
            if queue.empty() and concurrent_stream_processor.is_done():
                # all partitions were generated and processed. we're done here
                break

    def _handle_item(
        self,
        queue_item: QueueItem,
        concurrent_stream_processor: ConcurrentReadProcessor,
    ) -> Iterable[AirbyteMessage]:
        # handle queue item and call the appropriate handler depending on the type of the queue item
        if isinstance(queue_item, StreamThreadException):
            yield from concurrent_stream_processor.on_exception(queue_item)
        elif isinstance(queue_item, PartitionGenerationCompletedSentinel):
            yield from concurrent_stream_processor.on_partition_generation_completed(queue_item)
        elif isinstance(queue_item, Partition):
            concurrent_stream_processor.on_partition(queue_item)
        elif isinstance(queue_item, PartitionCompleteSentinel):
            yield from concurrent_stream_processor.on_partition_complete_sentinel(queue_item)
        elif isinstance(queue_item, Record):
            yield from concurrent_stream_processor.on_record(queue_item)
        elif isinstance(queue_item, AirbyteMessage):
            yield queue_item
        else:
            raise ValueError(f"Unknown queue item type: {type(queue_item)}")



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import base64
import json
from dataclasses import InitVar, dataclass
from datetime import datetime
from typing import Any, Mapping, MutableMapping, Optional, Union, cast

import jwt
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey
from cryptography.hazmat.primitives.asymmetric.ed448 import Ed448PrivateKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

from airbyte_cdk.sources.declarative.auth.declarative_authenticator import DeclarativeAuthenticator
from airbyte_cdk.sources.declarative.interpolation.interpolated_boolean import InterpolatedBoolean
from airbyte_cdk.sources.declarative.interpolation.interpolated_mapping import InterpolatedMapping
from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.declarative.requesters.request_option import (
    RequestOption,
    RequestOptionType,
)

# Type alias for keys that JWT library accepts
JwtKeyTypes = Union[
    RSAPrivateKey, EllipticCurvePrivateKey, Ed25519PrivateKey, Ed448PrivateKey, str, bytes
]


class JwtAlgorithm(str):
    """
    Enum for supported JWT algorithms
    """

    HS256 = "HS256"
    HS384 = "HS384"
    HS512 = "HS512"
    ES256 = "ES256"
    ES256K = "ES256K"
    ES384 = "ES384"
    ES512 = "ES512"
    RS256 = "RS256"
    RS384 = "RS384"
    RS512 = "RS512"
    PS256 = "PS256"
    PS384 = "PS384"
    PS512 = "PS512"
    EdDSA = "EdDSA"


@dataclass
class JwtAuthenticator(DeclarativeAuthenticator):
    """
    Generates a JSON Web Token (JWT) based on a declarative connector configuration file. The generated token is attached to each request via the Authorization header.

    Attributes:
        config (Mapping[str, Any]): The user-provided configuration as specified by the source's spec
        secret_key (Union[InterpolatedString, str]): The secret key used to sign the JWT
        algorithm (Union[str, JwtAlgorithm]): The algorithm used to sign the JWT
        token_duration (Optional[int]): The duration in seconds for which the token is valid
        base64_encode_secret_key (Optional[Union[InterpolatedBoolean, str, bool]]): Whether to base64 encode the secret key
        header_prefix (Optional[Union[InterpolatedString, str]]): The prefix to add to the Authorization header
        kid (Optional[Union[InterpolatedString, str]]): The key identifier to be included in the JWT header
        typ (Optional[Union[InterpolatedString, str]]): The type of the JWT.
        cty (Optional[Union[InterpolatedString, str]]): The content type of the JWT.
        iss (Optional[Union[InterpolatedString, str]]): The issuer of the JWT.
        sub (Optional[Union[InterpolatedString, str]]): The subject of the JWT.
        aud (Optional[Union[InterpolatedString, str]]): The audience of the JWT.
        additional_jwt_headers (Optional[Mapping[str, Any]]): Additional headers to include in the JWT.
        additional_jwt_payload (Optional[Mapping[str, Any]]): Additional payload to include in the JWT.
    """

    config: Mapping[str, Any]
    parameters: InitVar[Mapping[str, Any]]
    secret_key: Union[InterpolatedString, str]
    algorithm: Union[str, JwtAlgorithm]
    token_duration: Optional[int]
    base64_encode_secret_key: Optional[Union[InterpolatedBoolean, str, bool]] = False
    header_prefix: Optional[Union[InterpolatedString, str]] = None
    kid: Optional[Union[InterpolatedString, str]] = None
    typ: Optional[Union[InterpolatedString, str]] = None
    cty: Optional[Union[InterpolatedString, str]] = None
    iss: Optional[Union[InterpolatedString, str]] = None
    sub: Optional[Union[InterpolatedString, str]] = None
    aud: Optional[Union[InterpolatedString, str]] = None
    additional_jwt_headers: Optional[Mapping[str, Any]] = None
    additional_jwt_payload: Optional[Mapping[str, Any]] = None
    passphrase: Optional[Union[InterpolatedString, str]] = None
    request_option: Optional[RequestOption] = None

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        self._secret_key = InterpolatedString.create(self.secret_key, parameters=parameters)
        self._algorithm = (
            JwtAlgorithm(self.algorithm) if isinstance(self.algorithm, str) else self.algorithm
        )
        self._base64_encode_secret_key = (
            InterpolatedBoolean(self.base64_encode_secret_key, parameters=parameters)
            if isinstance(self.base64_encode_secret_key, str)
            else self.base64_encode_secret_key
        )
        self._token_duration = self.token_duration
        self._header_prefix = (
            InterpolatedString.create(self.header_prefix, parameters=parameters)
            if self.header_prefix
            else None
        )
        self._kid = InterpolatedString.create(self.kid, parameters=parameters) if self.kid else None
        self._typ = InterpolatedString.create(self.typ, parameters=parameters) if self.typ else None
        self._cty = InterpolatedString.create(self.cty, parameters=parameters) if self.cty else None
        self._iss = InterpolatedString.create(self.iss, parameters=parameters) if self.iss else None
        self._sub = InterpolatedString.create(self.sub, parameters=parameters) if self.sub else None
        self._aud = InterpolatedString.create(self.aud, parameters=parameters) if self.aud else None
        self._additional_jwt_headers = InterpolatedMapping(
            self.additional_jwt_headers or {}, parameters=parameters
        )
        self._additional_jwt_payload = InterpolatedMapping(
            self.additional_jwt_payload or {}, parameters=parameters
        )
        self._passphrase = (
            InterpolatedString.create(self.passphrase, parameters=parameters)
            if self.passphrase
            else None
        )

        # When we first implemented the JWT authenticator, we assumed that the signed token was always supposed
        # to be loaded into the request headers under the `Authorization` key. This is not always the case, but
        # this default option allows for backwards compatibility to be retained for existing connectors
        self._request_option = self.request_option or RequestOption(
            inject_into=RequestOptionType.header, field_name="Authorization", parameters=parameters
        )

    def _get_jwt_headers(self) -> dict[str, Any]:
        """
        Builds and returns the headers used when signing the JWT.
        """
        headers = self._additional_jwt_headers.eval(self.config, json_loads=json.loads)
        if any(prop in headers for prop in ["kid", "alg", "typ", "cty"]):
            raise ValueError(
                "'kid', 'alg', 'typ', 'cty' are reserved headers and should not be set as part of 'additional_jwt_headers'"
            )

        if self._kid:
            headers["kid"] = self._kid.eval(self.config, json_loads=json.loads)
        if self._typ:
            headers["typ"] = self._typ.eval(self.config, json_loads=json.loads)
        if self._cty:
            headers["cty"] = self._cty.eval(self.config, json_loads=json.loads)
        headers["alg"] = self._algorithm
        return headers

    def _get_jwt_payload(self) -> dict[str, Any]:
        """
        Builds and returns the payload used when signing the JWT.
        """
        now = int(datetime.now().timestamp())
        exp = now + self._token_duration if isinstance(self._token_duration, int) else now
        nbf = now

        payload = self._additional_jwt_payload.eval(self.config, json_loads=json.loads)
        if any(prop in payload for prop in ["iss", "sub", "aud", "iat", "exp", "nbf"]):
            raise ValueError(
                "'iss', 'sub', 'aud', 'iat', 'exp', 'nbf' are reserved properties and should not be set as part of 'additional_jwt_payload'"
            )

        if self._iss:
            payload["iss"] = self._iss.eval(self.config, json_loads=json.loads)
        if self._sub:
            payload["sub"] = self._sub.eval(self.config, json_loads=json.loads)
        if self._aud:
            payload["aud"] = self._aud.eval(self.config, json_loads=json.loads)

        payload["iat"] = now
        payload["exp"] = exp
        payload["nbf"] = nbf
        return payload

    def _get_secret_key(self) -> JwtKeyTypes:
        """
        Returns the secret key used to sign the JWT.
        """
        secret_key: str = self._secret_key.eval(self.config, json_loads=json.loads)

        if self._passphrase:
            passphrase_value = self._passphrase.eval(self.config, json_loads=json.loads)
            if passphrase_value:
                private_key = serialization.load_pem_private_key(
                    secret_key.encode(),
                    password=passphrase_value.encode(),
                )
                return cast(JwtKeyTypes, private_key)

        return (
            base64.b64encode(secret_key.encode()).decode()
            if self._base64_encode_secret_key
            else secret_key
        )

    def _get_signed_token(self) -> Union[str, Any]:
        """
        Signed the JWT using the provided secret key and algorithm and the generated headers and payload. For additional information on PyJWT see: https://pyjwt.readthedocs.io/en/stable/
        """
        try:
            return jwt.encode(
                payload=self._get_jwt_payload(),
                key=self._get_secret_key(),
                algorithm=self._algorithm,
                headers=self._get_jwt_headers(),
            )
        except Exception as e:
            raise ValueError(f"Failed to sign token: {e}")

    def _get_header_prefix(self) -> Union[str, None]:
        """
        Returns the header prefix to be used when attaching the token to the request.
        """
        return (
            self._header_prefix.eval(self.config, json_loads=json.loads)
            if self._header_prefix
            else None
        )

    @property
    def auth_header(self) -> str:
        options = self._get_request_options(RequestOptionType.header)
        return next(iter(options.keys()), "")

    @property
    def token(self) -> str:
        return (
            f"{self._get_header_prefix()} {self._get_signed_token()}"
            if self._get_header_prefix()
            else self._get_signed_token()
        )

    def get_request_params(self) -> Mapping[str, Any]:
        return self._get_request_options(RequestOptionType.request_parameter)

    def get_request_body_data(self) -> Union[Mapping[str, Any], str]:
        return self._get_request_options(RequestOptionType.body_data)

    def get_request_body_json(self) -> Mapping[str, Any]:
        return self._get_request_options(RequestOptionType.body_json)

    def _get_request_options(self, option_type: RequestOptionType) -> Mapping[str, Any]:
        options: MutableMapping[str, Any] = {}
        if self._request_option.inject_into == option_type:
            self._request_option.inject_into_request(options, self.token, self.config)
        return options



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass
from typing import Any, Iterable, Mapping, Optional

from airbyte_cdk.sources.declarative.partition_routers.partition_router import PartitionRouter
from airbyte_cdk.sources.types import StreamSlice, StreamState


@dataclass
class SinglePartitionRouter(PartitionRouter):
    """Partition router returning only a stream slice"""

    parameters: InitVar[Mapping[str, Any]]

    def get_request_params(
        self,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return {}

    def get_request_headers(
        self,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return {}

    def get_request_body_data(
        self,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return {}

    def get_request_body_json(
        self,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return {}

    def stream_slices(self) -> Iterable[StreamSlice]:
        yield StreamSlice(partition={}, cursor_slice={})

    def get_stream_state(self) -> Optional[Mapping[str, StreamState]]:
        """
        SinglePartitionRouter doesn't have parent streams
        """
        pass



import os

if os.getenv("DD_ENABLED", "false").lower() == "true":
    # Auto-instrumentation should be imported as early as possible.
    import ddtrace.auto  # noqa: F401

from fastapi import FastAPI

from .routers import capabilities, health, manifest

app = FastAPI(
    title="Manifest Server",
    description="A service for running low-code Airbyte connectors",
    version="0.2.0",
    contact={
        "name": "Airbyte",
        "url": "https://airbyte.com",
    },
)

app.include_router(health.router)
app.include_router(capabilities.router)
app.include_router(manifest.router, prefix="/v1")



#
# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass
from typing import Any, Dict, Mapping, Optional

from airbyte_cdk.sources.declarative.interpolation import InterpolatedString
from airbyte_cdk.sources.declarative.transformations import RecordTransformation
from airbyte_cdk.sources.types import Config, StreamSlice, StreamState


@dataclass
class KeysReplaceTransformation(RecordTransformation):
    """
    Transformation that applies keys names replacement.

    Example usage:
    - type: KeysReplace
      old: " "
      new: "_"
    Result:
    from: {"created time": ..., "customer id": ..., "user id": ...}
    to: {"created_time": ..., "customer_id": ..., "user_id": ...}
    """

    old: str
    new: str
    parameters: InitVar[Mapping[str, Any]]

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        self._old = InterpolatedString.create(self.old, parameters=parameters)
        self._new = InterpolatedString.create(self.new, parameters=parameters)

    def transform(
        self,
        record: Dict[str, Any],
        config: Optional[Config] = None,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
    ) -> None:
        if config is None:
            config = {}

        kwargs = {"record": record, "stream_state": stream_state, "stream_slice": stream_slice}
        old_key = str(self._old.eval(config, **kwargs))
        new_key = str(self._new.eval(config, **kwargs))

        def _transform(data: Dict[str, Any]) -> Dict[str, Any]:
            result = {}
            for key, value in data.items():
                updated_key = key.replace(old_key, new_key)
                if isinstance(value, dict):
                    result[updated_key] = _transform(value)
                else:
                    result[updated_key] = value
            return result

        transformed_record = _transform(record)
        record.clear()
        record.update(transformed_record)



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

from .add_fields import ConfigAddFields
from .remap_field import ConfigRemapField
from .remove_fields import ConfigRemoveFields

__all__ = ["ConfigRemapField", "ConfigAddFields", "ConfigRemoveFields"]



#
# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass
from typing import Any, Mapping, Optional, Union

from airbyte_cdk.sources.declarative.requesters.request_options.request_options_provider import (
    RequestOptionsProvider,
)
from airbyte_cdk.sources.types import StreamSlice, StreamState


@dataclass
class DefaultRequestOptionsProvider(RequestOptionsProvider):
    """
    Request options provider that extracts fields from the stream_slice and injects them into the respective location in the
    outbound request being made
    """

    parameters: InitVar[Mapping[str, Any]]

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        pass

    def get_request_params(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return {}

    def get_request_headers(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return {}

    def get_request_body_data(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Union[Mapping[str, Any], str]:
        return {}

    def get_request_body_json(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return {}
